cimport numpy as np
from libc.stddef cimport wchar_t


cdef extern from "Nd2ReadSdk.h":
    ctypedef void*           LIMFILEHANDLE
    ctypedef char            LIMCHAR
    ctypedef char*           LIMSTR
    ctypedef unsigned int    LIMUINT
    ctypedef int             LIMRESULT
    ctypedef size_t          LIMSIZE
    ctypedef bint            LIMBOOL
    ctypedef char*           LIMCSTR
    ctypedef const wchar_t*  LIMCWSTR

    LIMFILEHANDLE Lim_FileOpenForReadUtf8(LIMCSTR wszFileName)
    LIMFILEHANDLE Lim_FileOpenForRead(LIMCWSTR wszFileName)

    void      Lim_FileClose(LIMFILEHANDLE hFile)

    LIMSTR    Lim_FileGetAttributes(LIMFILEHANDLE hFile)
    LIMSTR    Lim_FileGetMetadata(LIMFILEHANDLE hFile)
    LIMSTR    Lim_FileGetFrameMetadata(LIMFILEHANDLE hFile, LIMUINT uiSeqIndex)
    LIMSTR    Lim_FileGetTextinfo(LIMFILEHANDLE hFile)
    LIMSTR    Lim_FileGetExperiment(LIMFILEHANDLE hFile)
    LIMUINT   Lim_FileGetSeqCount(LIMFILEHANDLE hFile)

    LIMSIZE   Lim_FileGetCoordSize(LIMFILEHANDLE hFile)
    LIMBOOL   Lim_FileGetSeqIndexFromCoords(LIMFILEHANDLE hFile, const LIMUINT * coords, LIMSIZE coordCount, LIMUINT* seqIdx)
    LIMSIZE   Lim_FileGetCoordsFromSeqIndex(LIMFILEHANDLE hFile, LIMUINT seqIdx, LIMUINT* coords, LIMSIZE maxCoordCount)
    LIMUINT   Lim_FileGetCoordInfo(LIMFILEHANDLE hFile, LIMUINT coord, LIMSTR type, LIMSIZE maxTypeSize)

    LIMRESULT Lim_FileGetImageData(LIMFILEHANDLE hFile, LIMUINT uiSeqIndex, LIMPICTURE* pPicture)

    LIMSIZE   Lim_InitPicture(LIMPICTURE* pPicture, LIMUINT width, LIMUINT height, LIMUINT bpc, LIMUINT components)
    void      Lim_DestroyPicture(LIMPICTURE* pPicture)

    void      Lim_FileFreeString(LIMSTR str)

    ctypedef struct LIMPICTURE:
        unsigned int  uiWidth             # !< Width (in pixels) of the picture
        unsigned int  uiHeight            # !< Height (in pixels) of the picture
        unsigned int  uiBitsPerComp       # !< Number of bits for each component
        unsigned int  uiComponents        # !< Number of components in each pixel
        size_t        uiWidthBytes        # !< Number of bytes for each pixel line (stride); aligned to 4bytes
        size_t        uiSize              # !< Number of bytes the image occupies
        void*         pImageData          # !< Image data


cdef inline LIMPICTURE nullpic():
    cdef LIMPICTURE p
    p.uiWidth = 0
    p.uiHeight = 0
    p.uiBitsPerComp = 0
    p.uiComponents = 0
    p.uiWidthBytes = 0
    p.uiSize = 0
    p.pImageData = NULL
    return p


ctypedef void (*destroyer_t)(LIMPICTURE*)


cdef class PicWrapper:
    cdef int dtype
    cdef LIMPICTURE pic
    cdef destroyer_t finalizer
    cdef void set_pic(PicWrapper self, LIMPICTURE pic, destroyer_t finalizer)
    cdef np.ndarray to_ndarray(self)



try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"
__author__ = "Talley Lambert"
__email__ = "talley.lambert@gmail.com"
__all__ = [
    "ND2File",
    "imread",
    "structures",
    "AXIS",
    "is_supported_file",
    "read_chunkmap",
    "rescue_nd2",
]


from . import structures
from ._chunkmap import read_chunkmap, rescue_nd2
from ._util import AXIS, is_supported_file
from .nd2file import ND2File, imread



from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np

from .. import structures

class ND2Reader:
    path: str
    def __init__(
        self,
        path: Union[str, Path],
        validate_frames: bool = False,
        search_window: int = 100,
        read_using_sdk: Optional[bool] = None,
    ) -> None: ...
    def open(self) -> None: ...
    def close(self) -> None: ...
    def __enter__(self) -> ND2Reader: ...
    def __exit__(self, *args) -> None: ...
    def _attributes(self) -> dict: ...
    @property
    def attributes(self) -> structures.Attributes: ...
    def _metadata(self) -> dict: ...
    def metadata(self) -> structures.Metadata: ...
    def _frame_metadata(self, seq_index: int) -> dict: ...
    def frame_metadata(self, seq_index: int) -> structures.FrameMetadata: ...
    def text_info(self) -> Dict[str, Any]: ...
    def _description(self) -> str: ...
    def _experiment(self) -> list: ...
    def experiment(self) -> List[structures.ExpLoop]: ...
    def sizes(self) -> dict: ...
    def _seq_count(self) -> int: ...
    def _coord_size(self) -> int: ...
    def _seq_index_from_coords(self, coords: Sequence) -> int: ...
    def _coords_from_seq_index(self, seq_index: int) -> Tuple[int, ...]: ...
    def _seq_index_from_pycoords(self, coords: Sequence) -> int: ...
    def _pycoords_from_seq_index(self, seq_index: int) -> Tuple[int, ...]: ...
    def _coord_info(self) -> List[Tuple[int, str, int]]: ...
    def _image(self, seq_index: int) -> np.ndarray: ...
    def voxel_size(self) -> Tuple[float, float, float]: ...
    def _custom_data(self) -> Dict[str, Any]: ...
    def _read_image(self, index: int) -> np.ndarray: ...
    def channel_names(self) -> List[str]: ...



import json
import mmap
import warnings
from pathlib import Path
from typing import List, Optional, Sequence, Tuple

import numpy as np

from .. import structures
from .._chunkmap import read_new_chunkmap

cimport numpy as np
from cpython cimport Py_INCREF, PyObject
from libc.stdint cimport uintptr_t
from libc.stdlib cimport free, malloc
from numpy cimport PyArray_SetBaseObject

np.import_array()

cdef class ND2Reader:

    cdef LIMFILEHANDLE _fh
    cdef public str path
    cdef bint _is_open
    cdef public dict _frame_map
    cdef public dict _meta_map
    cdef int _max_frame_index
    cdef _mmap
    cdef __strides
    cdef __attributes
    cdef __dtype
    cdef __raw_frame_shape
    cdef public _read_image
    cdef public _read_using_sdk
    cdef _wants_read_using_sdk

    def __cinit__(
        self,
        path: str | Path,
        validate_frames: bool = False,
        search_window: int = 100,
        read_using_sdk: Optional[bool] = None,
    ):
        self._is_open = 0
        self.__raw_frame_shape = None
        self._fh = NULL
        self.path = str(path)
        self._wants_read_using_sdk = read_using_sdk
        self.open()

        if read_using_sdk is None:
            read_using_sdk = self.attributes.compressionType is not None
        self._read_using_sdk = read_using_sdk

        if self._read_using_sdk:
            self._read_image = self._read_image_with_sdk
            self._frame_map, self._meta_map = {}, {}
            self._max_frame_index = 0
        else:
            self._read_image = self._read_image_from_memmap

            with open(path, 'rb') as pyfh:
                self._frame_map, self._meta_map = read_new_chunkmap(
                    pyfh, validate_frames=validate_frames, search_window=search_window
                )
            if validate_frames:
                self._frame_map = self._frame_map['good']

            self._max_frame_index = max(self._frame_map)


    cpdef open(self):
        if not self._is_open:
            self._fh = Lim_FileOpenForReadUtf8(self.path)
            if not self._fh:
                raise OSError("Could not open file: %s" % self.path)
            self._is_open = 1

            if self._wants_read_using_sdk is None:
                self._read_using_sdk = self.attributes.compressionType is not None
            else:
                self._read_using_sdk = self._wants_read_using_sdk
                if self.attributes.compressionType is not None and self._wants_read_using_sdk is False:
                    Lim_FileClose(self._fh)
                    raise ValueError("Cannot read compressed nd2 files with `read_using_sdk=False`")

            if not self._read_using_sdk:
                with open(self.path, 'rb') as fh:
                    self._mmap = mmap.mmap(fh.fileno(), 0, access=mmap.ACCESS_READ)

    cpdef close(self):
        if self._is_open:
            Lim_FileClose(self._fh)
            if not self._read_using_sdk:
                self._mmap.close()
            self._is_open = 0

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.close()

    cpdef dict _attributes(self):
        return _loads(Lim_FileGetAttributes(self._fh))

    @property
    def attributes(self) -> structures.Attributes:
        if not hasattr(self, '__attributes'):
            if not self._is_open:
                raise ValueError("Attempt to get attributes from closed nd2 file")
            cont = self._metadata().get('contents')
            attrs = self._attributes()
            nC = cont.get('channelCount') if cont else attrs.get("componentCount", 1)
            self.__attributes = structures.Attributes(**attrs, channelCount=nC)
        return self.__attributes

    def voxel_size(self) -> tuple[float, float, float]:
        meta = self.metadata()
        if meta:
            ch = meta.channels
            if ch:
                return ch[0].volume.axesCalibration
        return (1, 1, 1)

    def _metadata(self) -> dict:
        if not self._is_open:
            raise ValueError("Attempt to get metadata from closed nd2 file")
        return _loads(Lim_FileGetMetadata(self._fh))

    def metadata(self) -> structures.Metadata:
        return structures.Metadata(**self._metadata())

    def _frame_metadata(self, seq_index: int) -> dict:
        if not self._is_open:
            raise ValueError("Attempt to get frame_metadata from closed nd2 file")
        return _loads(Lim_FileGetFrameMetadata(self._fh, seq_index))

    def frame_metadata(self, seq_index: int) -> structures.Metadata:
        return structures.FrameMetadata(**self._frame_metadata(seq_index))

    def text_info(self) -> dict:
        if not self._is_open:
            raise ValueError("Attempt to get text_info from closed nd2 file")
        return _loads(Lim_FileGetTextinfo(self._fh))

    def _description(self) -> str:
        return self.text_info().get("description", '')

    def _experiment(self) -> list:
        if not self._is_open:
            raise ValueError("Attempt to get experiment from closed nd2 file")
        return _loads(Lim_FileGetExperiment(self._fh), list)

    def experiment(self) -> List[structures.ExpLoop]:
        from ..structures import _Loop
        return [_Loop.create(i) for i in self._experiment()]

    cpdef LIMUINT _seq_count(self):
        return Lim_FileGetSeqCount(self._fh)

    cpdef LIMSIZE _coord_size(self):
        return Lim_FileGetCoordSize(self._fh)

    def _seq_index_from_coords(self, coords: Sequence) -> int:
        if not self._is_open:
            raise ValueError("Attempt to seq_index from closed nd2 file")
        cdef LIMSIZE size = self._coord_size()
        if size == 0:
            return -1

        cdef LIMSIZE n = len(coords)
        if n != size:
            raise ValueError("Coords must be length: %d" % size)

        cdef LIMUINT seq_index = -1
        cdef LIMUINT *_coords
        _coords = <LIMUINT *>malloc(n * sizeof(LIMUINT))
        if not _coords:
            raise MemoryError()
        for i in range(n):
            _coords[i] = coords[i]

        try:
            if not Lim_FileGetSeqIndexFromCoords(self._fh, &_coords[0], n, &seq_index):
                raise ValueError("Coordinate %r has no sequence index" % (coords,))
            return seq_index
        finally:
            free(_coords)

    def _coords_from_seq_index(self, seq_index: int) -> tuple:
        cdef LIMSIZE size = self._coord_size()
        if size == 0:
            return ()

        cdef LIMUINT *output = <LIMUINT *> malloc(size * sizeof(LIMUINT))
        if not output:
            raise MemoryError()

        try:
            Lim_FileGetCoordsFromSeqIndex(self._fh, seq_index, output, size)
            return tuple([x for x in output[:size]])
        finally:
            free(output)

    def _coord_info(self) -> List[Tuple[int, str, int]]:
        cdef LIMCHAR loop_type[256]
        cdef LIMSIZE size = self._coord_size()
        if size == 0:
            return []

        out = []
        for i in range(size):
            loop_size = Lim_FileGetCoordInfo(self._fh, i, loop_type, 256)
            out.append((i, loop_type, loop_size))
        return out

    cdef _validate_seq(self, LIMUINT seq_index):
        cdef LIMUINT seq_count = self._seq_count()
        if seq_index >= seq_count:
            raise IndexError(
                "Sequence %d out of range (sequence count: %d)" % (seq_index, seq_count)
            )

    def _custom_data(self) -> dict:
        from .._xml import parse_xml_block

        return {
            k[14:]: parse_xml_block(self._get_meta_chunk(k))
            for k, v in self._meta_map.items()
            if k.startswith("CustomDataVar|")
        }

    def _get_meta_chunk(self, key: str) -> bytes:
        from .._chunkmap import read_chunk

        try:
            pos = self._meta_map[key]
        except KeyError:
            raise KeyError(
                f"No metdata chunk with key {key}. "
                f"Options include {set(self._meta_map)}"
            )
        with open(self.path, 'rb') as fh:
            return read_chunk(fh, pos)

    cdef _raw_frame_shape(self):
        if self.__raw_frame_shape is None:
            attr = self.attributes
            self.__raw_frame_shape = (
                    attr.heightPx,
                    attr.widthPx or -1,
                    attr.channelCount or 1,
                    attr.componentCount // (attr.channelCount or 1),
                )
        return self.__raw_frame_shape

    cdef _dtype(self):
        if self.__dtype is None:
            a = self.attributes
            d = a.pixelDataType[0] if a.pixelDataType else "u"
            self.__dtype = np.dtype(f"{d}{a.bitsPerComponentInMemory // 8}")
        return self.__dtype

    def _read_image_with_sdk(self, LIMUINT seq_index):
        self._validate_seq(seq_index)

        cdef LIMPICTURE pic = nullpic()
        cdef LIMRESULT result = Lim_FileGetImageData(self._fh, seq_index, &pic)

        if result != 0:
            error = LIM_ERR_CODE[result]
            raise RuntimeError('Error retrieving image data: %s' % error)

        array_wrapper = PicWrapper()
        array_wrapper.set_pic(pic, Lim_DestroyPicture)
        return array_wrapper.to_ndarray()

    cpdef np.ndarray _read_image_from_memmap(self, index: int):
        """Read a chunk directly without using SDK"""
        if index > self._max_frame_index:
            raise IndexError(f"Frame out of range: {index}")
        if not self._is_open:
            raise ValueError("Attempt to read from closed nd2 file")
        offset = self._frame_map.get(index, None)
        if offset is None:
            return self._missing_frame(index)

        # try:
        #     return np.ndarray(
        #         shape=self._raw_frame_shape(),
        #         dtype=self._dtype(),
        #         buffer=self._mmap,
        #         offset=offset,
        #         strides=self._strides,
        #     )
        # except TypeError:
        #     # If the chunkmap is wrong, and the mmap isn't long enough
        #     # for the requested offset & size, a TypeError is raised.
        #     return self._missing_frame(index)

        try:
            return np.frombuffer(
                self._mmap,
                dtype=self._dtype(),
                count=np.prod(self._raw_frame_shape()),
                offset=offset
            )  # this will be reshaped in nd2file.py

        except ValueError:
            # If the chunkmap is wrong, and the mmap isn't long enough
            # for the requested offset & size, a ValueError is raised.
            return self._missing_frame(index)

    @property
    def _strides(self):
        if not hasattr(self, '__strides'):
            a = self.attributes
            width = a.widthPx
            widthB = a.widthBytes
            if not (width and widthB):
                self.__strides = None
            else:
                bypc = a.bitsPerComponentInMemory // 8
                array_stride = widthB - (bypc * width * a.componentCount)
                if array_stride == 0:
                    self.__strides = None
                else:
                    self.__strides = (
                        array_stride + width * bypc * a.componentCount,
                        a.componentCount * bypc,
                        a.componentCount // a.channelCount * bypc,
                        bypc,
                    )
        return self.__strides

    cdef np.ndarray _missing_frame(self, int index = 0):
        # TODO: add other modes for filling missing data
        return np.zeros(self._raw_frame_shape(), self._dtype())

    cpdef channel_names(self):
        return [c.channel.name for c in self.metadata().channels or []]

    cpdef time_stamps(self):
        stamps = []
        for s in range(self._seq_count()):
            chs = self._frame_metadata(s).get('channels')
            if not chs:
                break
            stamps.append(chs[0]['time']['relativeTimeMs'])
        return stamps



cdef _loads(LIMSTR string, default=dict):
    if not string:
        return default()
    try:
        return json.loads(string)
    finally:
        Lim_FileFreeString(string)



LIM_ERR_CODE = {
    0: 'LIM_OK',
    -1: 'LIM_ERR_UNEXPECTED',
    -2: 'LIM_ERR_NOTIMPL',  # NotImplementedError
    -3: 'LIM_ERR_OUTOFMEMORY',  # MemoryError
    -4: 'LIM_ERR_INVALIDARG',
    -5: 'LIM_ERR_NOINTERFACE',
    -6: 'LIM_ERR_POINTER',
    -7: 'LIM_ERR_HANDLE',
    -8: 'LIM_ERR_ABORT',
    -9: 'LIM_ERR_FAIL',
    -10: 'LIM_ERR_ACCESSDENIED',
    -11: 'LIM_ERR_OS_FAIL',  # OSError
    -12: 'LIM_ERR_NOTINITIALIZED',
    -13: 'LIM_ERR_NOTFOUND',
    -14: 'LIM_ERR_IMPL_FAILED',
    -15: 'LIM_ERR_DLG_CANCELED',
    -16: 'LIM_ERR_DB_PROC_FAILED',
    -17: 'LIM_ERR_OUTOFRANGE',  # IndexError
    -18: 'LIM_ERR_PRIVILEGES',
    -19: 'LIM_ERR_VERSION',
}





# based on https://gist.github.com/GaelVaroquaux/1249305
# from Gael Varoquaux License: BSD 3 clause
cdef class PicWrapper:

    cdef void set_pic(PicWrapper self, LIMPICTURE pic, destroyer_t finalizer):
        if pic.uiBitsPerComp == 8:
            self.dtype = np.NPY_UINT8
        elif 8 < pic.uiBitsPerComp <= 16:
            self.dtype = np.NPY_UINT16
        elif pic.uiBitsPerComp == 32:
            self.dtype = np.NPY_UINT32
        else:
            raise ValueError("Unexpected bits per component: %d" % pic.uiBitsPerComp)

        self.pic = pic
        self.finalizer = finalizer

    def __array__(self):
        cdef np.npy_intp shape[3]
        shape[0] = <np.npy_intp> self.pic.uiHeight
        shape[1] = <np.npy_intp> self.pic.uiWidth
        shape[2] = <np.npy_intp> self.pic.uiComponents
        return np.PyArray_SimpleNewFromData(3, shape, self.dtype, self.pic.pImageData)

    def __dealloc__(self):
        self.finalizer(&self.pic)

    cdef np.ndarray to_ndarray(self):
        cdef np.ndarray ndarray = np.array(self, copy=False)

        # Assign our object to the 'base' of the ndarray object
        PyArray_SetBaseObject(ndarray, self)

        # Increment the reference count, as the above assignement was done in
        # C, and Python does not know that there is this additional reference
        Py_INCREF(self)
        return ndarray



import re
from datetime import datetime
from typing import IO, TYPE_CHECKING, Any, Callable, NamedTuple, Optional, Union

if TYPE_CHECKING:
    from os import PathLike

    from ._legacy import LegacyND2Reader
    from ._sdk.latest import ND2Reader

    StrOrBytesPath = Union[str, bytes, PathLike[str], PathLike[bytes]]


NEW_HEADER_MAGIC = b"\xda\xce\xbe\n"
OLD_HEADER_MAGIC = b"\x00\x00\x00\x0c"
VERSION = re.compile(r"^ND2 FILE SIGNATURE CHUNK NAME01!Ver([\d\.]+)$")


def is_supported_file(
    path: "StrOrBytesPath", open_: Callable[["StrOrBytesPath", str], IO[Any]] = open
):
    """Return `True` if `path` can be opened as an nd2 file.

    Parameters
    ----------
    path : Union[str, bytes, PathLike]
        A path to query
    open_ : Callable[[StrOrBytesPath, str], IO[Any]]
        Filesystem opener, by default `builtins.open`

    Returns
    -------
    bool
        Whether the can be opened.
    """
    with open_(path, "rb") as fh:
        return fh.read(4) in (NEW_HEADER_MAGIC, OLD_HEADER_MAGIC)


def get_reader(
    path: str,
    validate_frames: bool = False,
    search_window: int = 100,
    read_using_sdk: Optional[bool] = None,
) -> Union["ND2Reader", "LegacyND2Reader"]:
    with open(path, "rb") as fh:
        magic_num = fh.read(4)
        if magic_num == NEW_HEADER_MAGIC:
            from ._sdk.latest import ND2Reader

            return ND2Reader(
                path,
                validate_frames=validate_frames,
                search_window=search_window,
                read_using_sdk=read_using_sdk,
            )
        elif magic_num == OLD_HEADER_MAGIC:
            from ._legacy import LegacyND2Reader

            return LegacyND2Reader(path)
        raise OSError(
            f"file {path} not recognized as ND2.  First 4 bytes: {magic_num!r}"
        )


def is_new_format(path: str) -> bool:
    # TODO: this is just for dealing with missing test data
    with open(path, "rb") as fh:
        return fh.read(4) == NEW_HEADER_MAGIC


def jdn_to_datetime_local(jdn):
    return datetime.fromtimestamp((jdn - 2440587.5) * 86400.0)


def jdn_to_datetime_utc(jdn):
    return datetime.utcfromtimestamp((jdn - 2440587.5) * 86400.0)


def rgb_int_to_tuple(rgb):
    return ((rgb & 255), (rgb >> 8 & 255), (rgb >> 16 & 255))


DIMSIZE = re.compile(r"(\w+)'?\((\d+)\)")


def dims_from_description(desc) -> dict:
    if not desc:
        return {}
    match = re.search(r"Dimensions:\s?([^\r]+)\r?\n", desc)
    if not match:
        return {}
    dims = match.groups()[0]
    dims = dims.replace("λ", AXIS.CHANNEL)
    dims = dims.replace("XY", AXIS.POSITION)
    return {k: int(v) for k, v in DIMSIZE.findall(dims)}


class AXIS:
    X = "X"
    Y = "Y"
    Z = "Z"
    CHANNEL = "C"
    RGB = "S"
    TIME = "T"
    POSITION = "P"
    UNKNOWN = "U"

    _MAP = {
        "Unknown": UNKNOWN,
        "TimeLoop": TIME,
        "XYPosLoop": POSITION,
        "ZStackLoop": Z,
        "NETimeLoop": TIME,
    }


class VoxelSize(NamedTuple):
    x: float
    y: float
    z: float



import re
import struct
import threading
from pathlib import Path
from typing import BinaryIO, DefaultDict, Dict, List, Optional, Tuple, Union

import numpy as np

from . import structures as strct
from ._util import VoxelSize, dims_from_description
from ._xml import parse_xml_block

try:
    from functools import cached_property
except ImportError:
    cached_property = property  # type: ignore


I4s = struct.Struct(">I4s")
JP2_MAP_CHUNK = struct.Struct("<4s4sI")
IHDR = struct.Struct(">iihBB")  # yxc-dtype in jpeg 2000


class LegacyND2Reader:
    _fh: BinaryIO

    def __init__(self, path: Union[Path, str]):
        self._path = str(path)
        self._fh = open(self._path, mode="rb")
        length, box_type = I4s.unpack(self._fh.read(I4s.size))
        if length != 12 and box_type == b"jP  ":
            raise ValueError("File not recognized as Legacy ND2 (JPEG2000) format.")
        self._chunkmap = legacy_nd2_chunkmap(self._fh)
        self.lock = threading.RLock()

    def open(self) -> None:
        if self.closed:
            self._fh = open(self._path, mode="rb")

    def close(self) -> None:
        if not self.closed:
            self._fh.close()

    @property
    def closed(self) -> bool:
        return self._fh.closed

    def __enter__(self) -> "LegacyND2Reader":
        self.open()
        return self

    def __exit__(self, *_) -> None:
        self.close()

    @cached_property
    def ddim(self) -> dict:
        return dims_from_description(self.text_info().get("description"))

    def experiment(self) -> List[strct.ExpLoop]:
        meta = self._metadata
        exp = []
        if "LoopNo00" in meta:
            # old style:
            for i, (k, v) in enumerate(meta.items()):
                if k != "Version":
                    loop = self._make_loop(v, i)
                    if loop:
                        exp.append(loop)
        else:
            i = 0
            while meta:
                # ugh... another hack for weird metadata
                if len(meta) == 1:
                    meta = list(meta.values())[0]
                    if meta and isinstance(meta, list):
                        meta = meta[0]
                loop = self._make_loop(meta, i)
                if loop:
                    exp.append(loop)
                meta = meta.get("NextLevelEx")  # type: ignore
                i += 1
        return exp

    def _coord_info(self) -> List[Tuple[int, str, int]]:
        return [(i, l.type, l.count) for i, l in enumerate(self.experiment())]

    def _make_loop(
        self, meta_level: dict, nest_level: int = 0
    ) -> Optional[strct.ExpLoop]:
        """converts an old style metadata loop dict to a new ExpLoop structure."""
        type_ = meta_level.get("Type")
        params: dict = meta_level["LoopPars"]
        if type_ == 2:  # XYPosLoop
            poscount = len(params["PosX"])
            # empirically, it appears that some files list more positions in the
            # metadata than are actually recorded in the textinfo -> description.
            # NIS viewer seems to agree more with the description
            count = self.ddim.get("S") or params["Count"]
            points = []
            for i in range(count):
                idx = f"{poscount-i-1:05}"
                points.append(
                    strct.Position(
                        pfsOffset=params["PFSOffset"][idx],
                        stagePositionUm=strct.StagePosition(
                            x=params["PosX"][idx],
                            y=params["PosY"][idx],
                            z=params["PosZ"][idx],
                        ),
                    )
                )
            return strct.XYPosLoop(
                nestingLevel=nest_level,
                count=count,
                parameters=strct.XYPosLoopParams(
                    isSettingZ=params["UseZ"], points=points
                ),
            )
        if type_ == 8:  # TimeLoop
            # XXX strangely, I've seen files that seem to have a mult-phase
            # NETimeLoop, but use one of the Period...
            # try to find one that matches the dims in the description
            count = self.ddim.get("T")
            per = None
            if count is not None:
                for p in params["Period"].values():
                    if p["Count"] == count:
                        per = p
            # otherwise take the first period
            # XXX: this is definitely error prone.
            if per is None:
                per = next(iter(params["Period"].values()))
            return strct.TimeLoop(
                count=per["Count"],
                nestingLevel=nest_level,
                parameters=strct.TimeLoopParams(
                    startMs=per["Start"],
                    periodMs=per["Period"],
                    durationMs=per["Duration"],
                    periodDiff=strct.PeriodDiff(
                        avg=per.get("AvgPeriodDiff"),
                        max=per.get("MaxPeriodDiff"),
                        min=per.get("MinPeriodDiff"),
                    ),
                ),
            )
        if type_ == 4:
            return strct.ZStackLoop(
                count=params["Count"],
                nestingLevel=nest_level,
                parameters=strct.ZStackLoopParams(
                    homeIndex=params["ZHome"],
                    stepUm=params["ZStep"],
                    bottomToTop=params["ZLow"] < params["ZHigh"],
                ),
            )
        if type_ == 6:  # channel
            return None

        raise ValueError(f"unrecognized type: {type_}")

    @cached_property
    def attributes(self) -> strct.Attributes:
        head = self.header
        bpcim = head["bits_per_component"]
        bpcs = self._advanced_image_attributes.get("SignificantBits", bpcim)
        widthPx = head.get("columns")

        try:
            picplanes = self._get_xml_dict(b"VIMD", 0)["PicturePlanes"]
            nC = picplanes["Count"]
            compCount = picplanes["CompCount"]
        except Exception:
            compCount = nC = 1

        return strct.Attributes(
            bitsPerComponentInMemory=bpcim,
            bitsPerComponentSignificant=bpcs,
            componentCount=compCount,
            heightPx=head.get("rows", -1),
            widthPx=widthPx,
            widthBytes=widthPx * bpcim // 8 * compCount,
            pixelDataType="unsigned",
            sequenceCount=len(self._chunkmap[b"VCAL"]),
            compressionLevel=head.get("compression"),
            channelCount=nC,
        )

    def _get_xml_dict(self, key: bytes, index=0) -> dict:
        try:
            bxml = self._read_chunk(self._chunkmap[key][index])
            return parse_xml_block(bxml)
        except KeyError:
            return {}

    @cached_property
    def events(self) -> dict:
        return self._get_xml_dict(b"IEVE")

    # def sizes(self):
    #     attrs = cast(Attributes, self.attributes)

    def text_info(self) -> dict:
        d = self._get_xml_dict(b"TINF")
        for i in d.get("TextInfoItem", []):
            txt = i.get("Text", "")
            if txt.startswith("Metadata:"):
                return {"description": txt}
        return {}

    @cached_property
    def _advanced_image_attributes(self) -> dict:
        return self._get_xml_dict(b"ARTT").get("AdvancedImageAttributes", {})

    @cached_property
    def _metadata(self) -> dict:
        meta = self._get_xml_dict(b"AIM1") or self._get_xml_dict(b"AIMD")
        version = ""
        meta.pop("UnknownData", None)
        while len(meta) == 1:
            key, val = meta.popitem()
            if "_V" in key:
                version = key.split("_V")[1]
            meta = val
            meta.pop("UnknownData", None)
        meta["Version"] = version
        return meta

    def metadata(self) -> dict:
        return self._metadata

    @property
    def calibration(self) -> dict:
        return self._get_xml_dict(b"ACAL")

    def _read_chunk(self, pos) -> bytes:
        with self.lock:
            self._fh.seek(pos)
            length, box_type = I4s.unpack(self._fh.read(I4s.size))
            return self._fh.read(length - I4s.size)

    def _read_image(self, index: int) -> np.ndarray:
        try:
            from imagecodecs import jpeg2k_decode
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                f"{e}\n"
                f"Reading legacy format nd2 {self._fh.name!r} requires imagecodecs.\n"
                "Please install with `pip install imagecodecs`."
            )

        data = []
        cc = self.attributes.channelCount or 1
        for i in range(cc):
            d = self._read_chunk(self._chunkmap[b"LUNK"][index * cc + i])
            data.append(jpeg2k_decode(d))
        return np.stack(data, axis=-1)

    def frame_metadata(self, index: int) -> dict:
        return {
            **self._get_xml_dict(b"VCAL", index),
            **self._get_xml_dict(b"VIMD", index),
        }

    def _scan_vimd(self):
        zs = set()
        xys = set()
        ts = set()
        cs = set()
        for i in range(len(self._chunkmap[b"VIMD"])):
            xml = self._get_xml_dict(b"VIMD", i)
            ts.add(xml["TimeMSec"])
            xys.add((xml["XPos"], xml["YPos"]))
            for p in xml["PicturePlanes"]["Plane"].values():
                cs.add(p["OpticalConfigName"])
                zs.add(p["OpticalConfigFull"]["ZPosition0"])
        return (zs, xys, ts, cs)

    def voxel_size(self) -> VoxelSize:

        z: Optional[float] = None
        d = self.text_info().get("description") or ""
        _z = re.search(r"Z Stack Loop: 5\s+-\s+Step\s+([.\d]+)", d)
        if _z:
            try:
                z = float(_z.groups()[0])
            except Exception:
                pass
        if z is None:
            for e in self.experiment():
                if e.type == "ZStackLoop":
                    z = e.parameters.stepUm
                    break
        xy = self._get_xml_dict(b"VIMD", 0).get("Calibration") or 1
        return VoxelSize(xy, xy, z or 1)

    def channel_names(self) -> List[str]:
        xml = self._get_xml_dict(b"VIMD", 0)
        return [p["OpticalConfigName"] for p in xml["PicturePlanes"]["Plane"].values()]

    def time_stamps(self) -> List[str]:
        xml = self._get_xml_dict(b"VIMD", 0)
        return [p["OpticalConfigName"] for p in xml["PicturePlanes"]["Plane"].values()]

    @cached_property
    def header(self) -> dict:
        try:
            pos = self._chunkmap[b"jp2h"][0]
        except (KeyError, IndexError):
            raise KeyError("No valid jp2h header found in file")
        self._fh.seek(pos + I4s.size + 4)  # 4 bytes for "label"
        if self._fh.read(4) != b"ihdr":
            raise KeyError("No valid ihdr header found in jp2h header")

        w, h, c, t, z = IHDR.unpack(self._fh.read(IHDR.size))
        return {
            "rows": w,
            "columns": h,
            "channels": c,
            "bits_per_component": t + 1,
            "compression": z,
        }

    def _custom_data(self) -> dict:
        return {}


def legacy_nd2_chunkmap(fh: BinaryIO) -> Dict[bytes, List[int]]:
    fh.seek(-40, 2)
    sig, map_start = struct.unpack("<32sQ", fh.read())
    assert sig == b"LABORATORY IMAGING ND BOX MAP 00", "Not a legacy ND2 file"
    fh.seek(-map_start, 2)
    n_chunks = int.from_bytes(fh.read(4), "big")
    data = fh.read()
    d: DefaultDict[bytes, List[int]] = DefaultDict(list)
    for i in range(n_chunks):
        box_type, lim_type, offset = JP2_MAP_CHUNK.unpack_from(data, i * 16)
        if box_type in {b"jP  ", b"ftyp", b"jp2h"}:
            d[box_type].append(offset)
        else:
            d[lim_type].append(offset)
    return dict(d)


# def jp2_chunkmap(fh: io.BufferedReader) -> Dict[bytes, List[Tuple[int, int]]]:
#     """Retrieve chunk positions and shape from jpeg 2000 (legacy ND2) format

#     https://www.file-recovery.com/jp2-signature-format.htm

#     ISO/IEC 15444

#     The JPEG 2000 file format specification was based on the QuickTime
#     container format specification. A JPEG 2000 file is always big-endian.

#     JPEG 2000 files consist of consecutive chunks. Each chunk has 8 byte header:
#     4-byte chunk size (big-endian, high byte first) and 4-byte chunk type -
#     one of pre-defined signatures: "jP " or "jP2 ".

#     """
#     out = DefaultDict(list)
#     for box_type, pos, length in iter_jp2_chunks(fh):
#         out[box_type].append((pos, length))
#     return dict(out)


# def iter_jp2_chunks(fh: io.BufferedReader) -> Iterator[Tuple[bytes, int, int]]:
#     file_size = fh.seek(0, 2)
#     fh.seek(0)
#     pos = 0
#     while True:
#         length, box_type = I4s.unpack(fh.read(I4s.size))
#         yield box_type, pos, length
#         pos += length
#         if pos >= file_size:
#             break
#         fh.seek(pos)  # jump to next box



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import List, NamedTuple, Optional, Tuple, Union

from typing_extensions import Literal

# enums


class AxisInterpretation(str, Enum):
    distance = "distance"
    time = "time"


# tuples


class Attributes(NamedTuple):
    bitsPerComponentInMemory: int
    bitsPerComponentSignificant: int
    componentCount: int
    heightPx: int
    pixelDataType: str
    sequenceCount: int
    widthBytes: Optional[int] = None
    widthPx: Optional[int] = None
    compressionLevel: Optional[int] = None
    compressionType: Optional[str] = None
    tileHeightPx: Optional[int] = None
    tileWidthPx: Optional[int] = None
    channelCount: Optional[int] = None


class ImageInfo(NamedTuple):
    width: int
    height: int
    components: int
    bits_per_pixel: int


# experiment #################


LoopTypeString = Union[
    Literal["TimeLoop"],
    Literal["NETimeLoop"],
    Literal["XYPosLoop"],
    Literal["ZStackLoop"],
]


@dataclass
class _Loop:
    count: int
    nestingLevel: int
    parameters: LoopParams
    type: LoopTypeString

    @classmethod
    def create(cls, obj: dict) -> ExpLoop:
        return globals()[obj["type"]](**obj)


#####


@dataclass
class TimeLoop(_Loop):
    parameters: TimeLoopParams
    type: Literal["TimeLoop"] = "TimeLoop"

    def __post_init__(self):
        if isinstance(self.parameters, dict):
            self.parameters = TimeLoopParams(**self.parameters)


@dataclass
class TimeLoopParams:
    startMs: float
    periodMs: float
    durationMs: float
    periodDiff: PeriodDiff

    def __post_init__(self):
        if isinstance(self.periodDiff, dict):
            self.periodDiff = PeriodDiff(**self.periodDiff)


@dataclass
class PeriodDiff:
    avg: float = 0
    max: float = 0
    min: float = 0


######


@dataclass
class NETimeLoop(_Loop):
    parameters: NETimeLoopParams
    type: Literal["NETimeLoop"] = "NETimeLoop"

    def __post_init__(self):
        if isinstance(self.parameters, dict):
            self.parameters = NETimeLoopParams(**self.parameters)


@dataclass
class NETimeLoopParams:
    periods: List[Period]

    def __post_init__(self):
        self.periods = [Period(**i) if isinstance(i, dict) else i for i in self.periods]


@dataclass
class Period(TimeLoopParams):
    count: int


#########


@dataclass
class XYPosLoop(_Loop):
    parameters: XYPosLoopParams
    type: Literal["XYPosLoop"] = "XYPosLoop"

    def __post_init__(self):
        if isinstance(self.parameters, dict):
            self.parameters = XYPosLoopParams(**self.parameters)


@dataclass
class XYPosLoopParams:
    isSettingZ: bool
    points: List[Position]

    def __post_init__(self):
        self.points = [Position(**i) if isinstance(i, dict) else i for i in self.points]


@dataclass
class Position:
    stagePositionUm: StagePosition
    pfsOffset: Optional[float] = None
    name: Optional[str] = None

    def __post_init__(self):
        if isinstance(self.stagePositionUm, dict):
            self.stagePositionUm = StagePosition(*self.stagePositionUm)


class StagePosition(NamedTuple):
    x: float
    y: float
    z: float


# ######


@dataclass
class ZStackLoop(_Loop):
    parameters: ZStackLoopParams
    type: Literal["ZStackLoop"] = "ZStackLoop"

    def __post_init__(self):
        if isinstance(self.parameters, dict):
            self.parameters = ZStackLoopParams(**self.parameters)


@dataclass
class ZStackLoopParams:
    homeIndex: int
    stepUm: float
    bottomToTop: bool
    deviceName: Optional[str] = None


###

ExpLoop = Union[TimeLoop, NETimeLoop, XYPosLoop, ZStackLoop]
LoopParams = Union[TimeLoopParams, NETimeLoopParams, XYPosLoopParams, ZStackLoopParams]

# metadata #################


@dataclass
class Metadata:
    contents: Optional[Contents] = None
    channels: Optional[List[Channel]] = None

    def __post_init__(self):
        if isinstance(self.contents, dict):
            self.contents = Contents(**self.contents)
        if self.channels:
            self.channels = [
                Channel(**i) if isinstance(i, dict) else i for i in self.channels
            ]


@dataclass
class Contents:
    channelCount: int
    frameCount: int


@dataclass
class Channel:
    channel: ChannelMeta
    loops: Optional[LoopIndices]
    microscope: Microscope
    volume: Volume

    def __post_init__(self):
        self.channel = ChannelMeta(**self.channel)
        self.microscope = Microscope(**self.microscope)
        self.volume = Volume(**self.volume)
        if self.loops:
            self.loops = LoopIndices(**self.loops)


@dataclass
class ChannelMeta:
    name: str
    index: int
    colorRGB: int  # probably 0xBBGGRR
    emissionLambdaNm: Optional[float] = None
    excitationLambdaNm: Optional[float] = None


@dataclass
class LoopIndices:
    NETimeLoop: Optional[int] = None
    TimeLoop: Optional[int] = None
    XYPosLoop: Optional[int] = None
    ZStackLoop: Optional[int] = None


@dataclass
class Microscope:
    objectiveMagnification: Optional[float] = None
    objectiveName: Optional[str] = None
    objectiveNumericalAperture: Optional[float] = None
    zoomMagnification: Optional[float] = None
    immersionRefractiveIndex: Optional[float] = None
    projectiveMagnification: Optional[float] = None
    pinholeDiameterUm: Optional[float] = None
    modalityFlags: List[str] = field(default_factory=list)


@dataclass
class Volume:
    axesCalibrated: Tuple[bool, bool, bool]
    axesCalibration: Tuple[float, float, float]
    axesInterpretation: Tuple[
        AxisInterpretation, AxisInterpretation, AxisInterpretation
    ]
    bitsPerComponentInMemory: int
    bitsPerComponentSignificant: int
    cameraTransformationMatrix: Tuple[float, float, float, float]
    componentCount: int
    componentDataType: Union[Literal["unsigned"], Literal["float"]]
    voxelCount: Tuple[int, int, int]
    componentMaxima: Optional[List[float]] = None
    componentMinima: Optional[List[float]] = None
    pixelToStageTransformationMatrix: Optional[
        Tuple[float, float, float, float, float, float]
    ] = None

    # NIS Microscope Absolute frame in um =
    # pixelToStageTransformationMatrix * (X_in_px,  Y_in_px,  1) + stagePositionUm

    def __post_init__(self):
        self.axesInterpretation = tuple(  # type: ignore
            AxisInterpretation(i) for i in self.axesInterpretation
        )


@dataclass
class TimeStamp:
    absoluteJulianDayNumber: float
    relativeTimeMs: float


@dataclass
class FrameChannel(Channel):
    position: Position
    time: TimeStamp

    def __post_init__(self):
        super().__post_init__()
        self.position = Position(**self.position)
        self.time = TimeStamp(**self.time)


@dataclass
class FrameMetadata:
    contents: Contents
    channels: List[FrameChannel]

    def __post_init__(self):
        self.contents = Contents(**self.contents)
        self.channels = [FrameChannel(**i) for i in self.channels]


class Coordinate(NamedTuple):
    index: int
    type: str
    size: int



from __future__ import annotations

import mmap
import threading
from enum import Enum
from itertools import product
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Optional,
    Sequence,
    Set,
    Sized,
    SupportsInt,
    Union,
    cast,
    overload,
)

import numpy as np

from ._util import AXIS, VoxelSize, get_reader, is_supported_file
from .structures import Attributes, ExpLoop, FrameMetadata, Metadata, XYPosLoop

try:
    from functools import cached_property
except ImportError:
    cached_property = property  # type: ignore


if TYPE_CHECKING:
    from typing import Any, Dict, List, Tuple

    import dask.array.core
    import xarray as xr
    from typing_extensions import Literal


Index = Union[int, slice]


class ReadMode(str, Enum):
    MMAP = "mmap"
    SDK = "sdk"


class ND2File:
    _memmap: mmap.mmap
    _is_legacy: bool

    def __init__(
        self,
        path: Union[Path, str],
        *,
        validate_frames: bool = False,
        search_window: int = 100,
        read_using_sdk: bool = None,
    ) -> None:
        """Open an nd2 file.

        Parameters
        ----------
        path : Union[Path, str]
            Filename of an nd2 file.
        validate_frames : bool
            Whether to verify (and attempt to fix) frames whose positions have been
            shifted relative to the predicted offset (i.e. in a corrupted file).
            This comes at a slight performance penalty at file open, but may "rescue"
            some corrupt files. by default False.
        search_window : int
            When validate_frames is true, this is the search window (in KB) that will
            be used to try to find the actual chunk position. by default 100 KB
        read_using_sdk : Optional[bool]
            If `True`, use the SDK to read the file. If `False`, inspects the chunkmap
            and reads from a `numpy.memmap`. If `None` (the default), uses the SDK if
            the file is compressed, otherwise uses the memmap. Note: using
            `read_using_sdk=False` on a compressed file will result in a ValueError.

        """
        self._path = str(path)
        self._rdr = get_reader(
            self._path,
            validate_frames=validate_frames,
            search_window=search_window,
            read_using_sdk=read_using_sdk,
        )
        self._closed = False
        self._is_legacy = "Legacy" in type(self._rdr).__name__
        self._lock = threading.RLock()

    @staticmethod
    def is_supported_file(path) -> bool:
        return is_supported_file(path)

    @property
    def path(self):
        """Path of the image."""
        return self._path

    @property
    def is_legacy(self) -> bool:
        """Whether file is a legacy nd2 (JPEG2000) file."""
        return self._is_legacy

    def open(self) -> None:
        """open file for reading."""
        if self.closed:
            self._rdr.open()
            self._closed = False

    def close(self) -> None:
        """Close file (may cause segfault if read when closed in some cases)."""
        if not self.closed:
            self._rdr.close()
            self._closed = True

    @property
    def closed(self) -> bool:
        """Whether the file is closed."""
        return self._closed

    def __enter__(self) -> ND2File:
        self.open()
        return self

    def __exit__(self, *_) -> None:
        self.close()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_rdr"]
        del state["_lock"]
        return state

    def __setstate__(self, d):
        self.__dict__ = d
        self._lock = threading.RLock()
        self._rdr = get_reader(self._path)
        if self._closed:
            self._rdr.close()

    @cached_property
    def attributes(self) -> Attributes:
        """Core image attributes"""
        return self._rdr.attributes

    @cached_property
    def text_info(self) -> Dict[str, Any]:
        """Misc text info."""
        return self._rdr.text_info()

    @cached_property
    def experiment(self) -> List[ExpLoop]:
        """Loop information for each nd axis"""
        return self._rdr.experiment()

    @cached_property
    def metadata(self) -> Union[Metadata, dict]:
        """Various metadata (will be dict if legacy format)."""
        return self._rdr.metadata()

    def frame_metadata(
        self, seq_index: Union[int, tuple]
    ) -> Union[FrameMetadata, dict]:
        """Metadata for specific frame.

        This includes the global metadata from the metadata function.
        (will be dict if legacy format).

        Parameters
        ----------
        seq_index : Union[int, tuple]
            frame index

        Returns
        -------
        Union[FrameMetadata, dict]
            dict if legacy format, else FrameMetadata
        """

        idx = cast(
            int,
            self._seq_index_from_coords(seq_index)
            if isinstance(seq_index, tuple)
            else seq_index,
        )
        return self._rdr.frame_metadata(idx)

    @cached_property
    def custom_data(self) -> Dict[str, Any]:
        """Dict of various unstructured custom metadata."""
        return self._rdr._custom_data()

    @cached_property
    def ndim(self) -> int:
        """number of dimensions"""
        return len(self.shape)

    @cached_property
    def shape(self) -> Tuple[int, ...]:
        """size of each axis"""
        return self._coord_shape + self._frame_shape

    @cached_property
    def sizes(self) -> Dict[str, int]:
        """names and sizes for each axis"""
        attrs = self.attributes
        dims = {AXIS._MAP[c[1]]: c[2] for c in self._rdr._coord_info()}
        dims[AXIS.CHANNEL] = (
            dims.pop(AXIS.CHANNEL)
            if AXIS.CHANNEL in dims
            else (attrs.channelCount or 1)
        )
        dims[AXIS.Y] = attrs.heightPx
        dims[AXIS.X] = attrs.widthPx or -1
        if self.components_per_channel == 3:  # rgb
            dims[AXIS.RGB] = self.components_per_channel
        else:
            # if not exactly 3 channels, throw them all into monochrome channels
            dims[AXIS.CHANNEL] = attrs.componentCount
        return {k: v for k, v in dims.items() if v != 1}

    @property
    def is_rgb(self) -> bool:
        """Whether the image is rgb"""
        return self.components_per_channel in (3, 4)

    @property
    def components_per_channel(self) -> int:
        """Number of components per channel (e.g. 3 for rgb)"""
        attrs = cast(Attributes, self.attributes)
        return attrs.componentCount // (attrs.channelCount or 1)

    @property
    def size(self) -> int:
        """Total number of pixels in the volume."""
        return int(np.prod(self.shape))

    @property
    def nbytes(self) -> int:
        """Total bytes of image data."""
        return self.size * self.dtype.itemsize

    @cached_property
    def dtype(self) -> np.dtype:
        """Image data type"""
        attrs = self.attributes
        d = attrs.pixelDataType[0] if attrs.pixelDataType else "u"
        return np.dtype(f"{d}{attrs.bitsPerComponentInMemory // 8}")

    def voxel_size(self, channel: int = 0) -> VoxelSize:
        """XYZ voxel size.

        Parameters
        ----------
        channel : int
            Channel for which to retrieve voxel info, by default 0

        Returns
        -------
        VoxelSize
            Named tuple with attrs `x`, `y`, and `z`.
        """
        return VoxelSize(*self._rdr.voxel_size())

    def asarray(self, position: Optional[int] = None) -> np.ndarray:
        """Read image into numpy array.

        Parameters
        ----------
        position : int, optional
            A specific XY position to extract, by default (None) reads all.

        Returns
        -------
        np.ndarray

        Raises
        ------
        ValueError
            if `position` is a string and is not a valid position name
        IndexError
            if `position` is provided and is out of range
        """
        final_shape = list(self.shape)
        if position is None:
            seqs: Sequence[int] = range(self._frame_count)
        else:
            if isinstance(position, str):
                try:
                    position = self._position_names().index(position)
                except ValueError as e:
                    raise ValueError(
                        f"{position!r} is not a valid position name"
                    ) from e
            try:
                pidx = list(self.sizes).index(AXIS.POSITION)
            except ValueError as exc:
                if position > 0:
                    raise IndexError(
                        f"Position {position} is out of range. "
                        f"Only 1 position available"
                    ) from exc
                seqs = range(self._frame_count)
            else:
                if position >= self.sizes[AXIS.POSITION]:
                    raise IndexError(
                        f"Position {position} is out of range. "
                        f"Only {self.sizes[AXIS.POSITION]} positions available"
                    )

                ranges: List[Union[range, tuple]] = [
                    range(x) for x in self._coord_shape
                ]
                ranges[pidx] = (position,)
                coords = list(zip(*product(*ranges)))
                seqs = self._seq_index_from_coords(coords)  # type: ignore
                final_shape[pidx] = 1

        arr: np.ndarray = np.stack([self._get_frame(i) for i in seqs])
        return arr.reshape(final_shape)

    def __array__(self) -> np.ndarray:
        """array protocol"""
        return self.asarray()

    def to_dask(self, wrapper=True, copy=True) -> dask.array.core.Array:
        """Create dask array (delayed reader) representing image.

        This generally works well, but it remains to be seen whether performance
        is optimized, or if we're duplicating safety mechanisms. You may try
        various combinations of `wrapper` and `copy`, setting both to `False`
        will very likely cause segmentation faults in many cases.  But setting
        one of them to `False`, may slightly improve read speed in certain
        cases.

        Parameters
        ----------
        wrapper : bool
            If True (the default), the returned obect will be a thin subclass of
            a :class:`dask.array.Array` (an
            `ResourceBackedDaskArray`) that manages the opening and closing of this file
            when getting chunks via compute(). If `wrapper` is `False`, then a pure
            `dask.array.core.Array` will be returned. However, when that array is
            computed, it will incur a file open/close on *every* chunk that is read (in
            the `_dask_block` method).  As such `wrapper` will generally be much faster,
            however, it *may* fail (i.e. result in segmentation faults) with certain
            dask schedulers.
        copy : bool
            If `True` (the default), the dask chunk-reading function will return
            an array copy. This can avoid segfaults in certain cases, though it
            may also add overhead.

        Returns
        -------
        dask.array.core.Array
        """
        from dask.array import map_blocks

        chunks = [(1,) * x for x in self._coord_shape]
        chunks += [(x,) for x in self._frame_shape]
        dask_arr = map_blocks(
            self._dask_block,
            copy=copy,
            chunks=chunks,
            dtype=self.dtype,
        )
        if wrapper:
            from resource_backed_dask_array import ResourceBackedDaskArray

            # this subtype allows the dask array to re-open the underlying
            # nd2 file on compute.
            return ResourceBackedDaskArray.from_array(dask_arr, self)
        return dask_arr

    _NO_IDX = -1

    def _seq_index_from_coords(
        self, coords: Sequence
    ) -> Union[Sequence[int], SupportsInt]:
        if not self._coord_shape:
            return self._NO_IDX
        return np.ravel_multi_index(coords, self._coord_shape)

    def _dask_block(self, copy: bool, block_id: Tuple[int]) -> np.ndarray:
        if isinstance(block_id, np.ndarray):
            return
        with self._lock:
            was_closed = self.closed
            if self.closed:
                self.open()
            try:
                ncoords = len(self._coord_shape)
                idx = self._seq_index_from_coords(block_id[:ncoords])

                if idx == self._NO_IDX:
                    if any(block_id):
                        raise ValueError(
                            f"Cannot get chunk {block_id} for single frame image."
                        )
                    idx = 0
                data = self._get_frame(cast(int, idx))
                data = data.copy() if copy else data
                return data[(np.newaxis,) * ncoords]
            finally:
                if was_closed:
                    self.close()

    def to_xarray(
        self,
        delayed: bool = True,
        squeeze: bool = True,
        position: Optional[int] = None,
        copy: bool = True,
    ) -> xr.DataArray:
        """Create labeled xarray representing image.

        `array.dims` will be populated according to image metadata, and coordinates
        will be populated based on pixel spacings. Additional metadata is available
        in `array.attrs['metadata']`.

        Parameters
        ----------
        delayed : bool
            Whether the DataArray should be backed by dask array or numpy array,
            by default True (dask).
        squeeze : bool
            Whether to squeeze singleton dimensions, by default True
        position : int, optional
            A specific XY position to extract, by default (None) reads all.
        copy : bool
            Only applies when `delayed==True`.  See `to_dask` for details.

        Returns
        -------
        xr.DataArray
            xarray with all axes labeled.
        """
        import xarray as xr

        data = self.to_dask(copy=copy) if delayed else self.asarray(position)
        dims = list(self.sizes)
        coords = self._expand_coords(squeeze)
        if not squeeze:
            for missing_dim in set(coords).difference(dims):
                dims.insert(0, missing_dim)
            missing_axes = len(dims) - data.ndim
            if missing_axes > 0:
                data = data[(np.newaxis,) * missing_axes]

        if position is not None and not delayed and AXIS.POSITION in coords:
            # if it's delayed, we do this using isel below instead.
            coords[AXIS.POSITION] = [coords[AXIS.POSITION][position]]

        x = xr.DataArray(
            data,
            dims=dims,
            coords=coords,
            attrs={
                "metadata": {
                    "metadata": self.metadata,
                    "experiment": self.experiment,
                    "attributes": self.attributes,
                    "text_info": self.text_info,
                }
            },
        )
        if delayed and position is not None and AXIS.POSITION in coords:
            x = x.isel({AXIS.POSITION: [position]})
        return x.squeeze() if squeeze else x

    @property
    def _frame_coords(self) -> Set[str]:
        return {AXIS.X, AXIS.Y, AXIS.CHANNEL, AXIS.RGB}

    @property
    def _raw_frame_shape(self) -> Tuple[int, int, int, int]:
        """sizes of each frame coordinate, prior to reshape"""
        attr = self.attributes
        return (
            attr.heightPx,
            attr.widthPx or -1,
            attr.channelCount or 1,
            self.components_per_channel,
        )

    @property
    def _frame_shape(self) -> Tuple[int, ...]:
        """sizes of each frame coordinate, after reshape & squeeze"""
        return tuple(v for k, v in self.sizes.items() if k in self._frame_coords)

    @cached_property
    def _coord_shape(self) -> Tuple[int, ...]:
        """sizes of each *non-frame* coordinate"""
        return tuple(v for k, v in self.sizes.items() if k not in self._frame_coords)

    @property
    def _frame_count(self) -> int:
        return int(np.prod(self._coord_shape))

    def _get_frame(self, index: int) -> np.ndarray:
        frame = self._rdr._read_image(index)
        frame.shape = self._raw_frame_shape
        return frame.transpose((2, 0, 1, 3)).squeeze()

    def _expand_coords(self, squeeze: bool = True) -> dict:
        """Return a dict that can be used as the coords argument to xr.DataArray

        Parameters
        ----------
        squeeze : bool
            whether to squeeze axes with length < 2, by default True

        Returns
        -------
        dict
            dict of axis name -> coordinates
        """
        dx, dy, dz = self.voxel_size()

        coords: Dict[str, Sized] = {
            AXIS.Y: np.arange(self.attributes.heightPx) * dy,
            AXIS.X: np.arange(self.attributes.widthPx or 1) * dx,
            AXIS.CHANNEL: self._channel_names,
            AXIS.POSITION: ["XYPos:0"],  # maybe overwritten below
        }

        for c in self.experiment:
            if squeeze and c.count <= 1:
                continue
            if c.type == "ZStackLoop":
                coords[AXIS.Z] = np.arange(c.count) * c.parameters.stepUm
            elif c.type == "TimeLoop":
                coords[AXIS.TIME] = np.arange(c.count) * c.parameters.periodMs
            elif c.type == "NETimeLoop":
                pers = [np.arange(p.count) * p.periodMs for p in c.parameters.periods]
                coords[AXIS.TIME] = np.hstack(pers)
            elif c.type == "XYPosLoop":
                coords[AXIS._MAP["XYPosLoop"]] = self._position_names(c)

        if self.components_per_channel > 1:
            coords[AXIS.RGB] = ["Red", "Green", "Blue", "alpha"][
                : self.components_per_channel
            ]

        # fix for Z axis missing from experiment:
        if AXIS.Z in self.sizes and AXIS.Z not in coords:
            coords[AXIS.Z] = np.arange(self.sizes[AXIS.Z]) * dz

        if squeeze:
            return {k: v for k, v in coords.items() if len(v) > 1}
        return coords

    def _position_names(self, loop: Optional[XYPosLoop] = None) -> List[str]:
        if loop is None:
            for c in self.experiment:
                if c.type == "XYPosLoop":
                    loop = c
                    break
        if loop is None:
            return ["XYPos:0"]
        return [p.name or f"XYPos:{i}" for i, p in enumerate(loop.parameters.points)]

    @property
    def _channel_names(self) -> List[str]:
        return self._rdr.channel_names()

    def __repr__(self) -> str:
        try:
            details = " (closed)" if self.closed else f" {self.dtype}: {self.sizes!r}"
            extra = f": {Path(self.path).name!r}{details}"
        except Exception:
            extra = ""
        return f"<ND2File at {hex(id(self))}{extra}>"


@overload
def imread(
    file: Union[Path, str],
    *,
    dask: Literal[False],
    xarray: Literal[False],
    validate_frames: bool = False,
    read_using_sdk: Optional[bool] = None,
) -> np.ndarray:
    ...


@overload
def imread(
    file: Union[Path, str],
    *,
    dask: bool = ...,
    xarray: Literal[True],
    validate_frames: bool = False,
    read_using_sdk: Optional[bool] = None,
) -> xr.DataArray:
    ...


@overload
def imread(
    file: Union[Path, str],
    *,
    dask: Literal[True],
    xarray: Literal[False],
    validate_frames: bool = False,
    read_using_sdk: Optional[bool] = None,
) -> dask.array.core.Array:
    ...


def imread(
    file: Union[Path, str],
    *,
    dask: bool = False,
    xarray: bool = False,
    validate_frames: bool = False,
    read_using_sdk: Optional[bool] = None,
):
    """Open `file`, return requested array type, and close `file`.

    Parameters
    ----------
    file : Union[Path, str]
        Filepath (`str`) or `Path` object to ND2 file.
    dask : bool
        If `True`, returns a (delayed) `dask.array.Array`. This will avoid reading
        any data from disk until specifically requested by using `.compute()` or
        casting to a numpy array with `np.asarray()`. By default `False`.
    xarray : bool
        If `True`, returns an `xarray.DataArray`, `array.dims` will be populated
        according to image metadata, and coordinates will be populated based on pixel
        spacings. Additional metadata is available in `array.attrs['metadata']`.
        If `dask` is also `True`, will return an xarray backed by a delayed dask array.
        By default `False`.
    validate_frames : bool
        Whether to verify (and attempt to fix) frames whose positions have been
        shifted relative to the predicted offset (i.e. in a corrupted file).
        This comes at a slight performance penalty at file open, but may "rescue"
        some corrupt files. by default False.
    read_using_sdk : Optional[bool]
        If `True`, use the SDK to read the file. If `False`, inspects the chunkmap and
        reads from a `numpy.memmap`. If `None` (the default), uses the SDK if the file
        is compressed, otherwise uses the memmap.
        Note: using `read_using_sdk=False` on a compressed file will result in a
        ValueError.

    Returns
    -------
    Union[np.ndarray, dask.array.Array, xarray.DataArray]
        Array subclass, depending on arguments used.
    """
    with ND2File(
        file, validate_frames=validate_frames, read_using_sdk=read_using_sdk
    ) as nd2:
        if xarray:
            return nd2.to_xarray(delayed=dask)
        elif dask:
            return nd2.to_dask()
        else:
            return nd2.asarray()



import re
from functools import partial
from typing import Any, Callable, Dict, Optional

import lxml.etree


def parse_xml_block(bxml: bytes) -> Dict[str, Any]:
    node = lxml.etree.XML(bxml.split(b"?>", 1)[-1])
    return elem2dict(node)


lower = re.compile("^[a-z_]+")
_TYPEMAP: Dict[Optional[str], Callable] = {
    "bool": bool,
    "lx_uint32": int,
    "lx_uint64": int,
    "lx_int32": int,
    "lx_int64": int,
    "double": float,
    "CLxStringW": str,
    "CLxByteArray": partial(bytes, encoding="utf-8"),
    "unknown": str,
    None: str,
}


def elem2dict(node: lxml.etree._Element) -> Dict[str, Any]:
    """
    Convert an lxml.etree node tree into a dict.
    """
    result: Dict[str, Any] = {}

    if "value" in node.attrib:
        type_ = _TYPEMAP[node.attrib.get("runtype")]
        try:
            return type_(node.attrib["value"])
        except ValueError:
            return node.attrib["value"]

    attrs = node.attrib
    attrs.pop("runtype", None)
    attrs.pop("version", None)
    result.update(node.attrib)

    for element in node.iterchildren():
        # Remove namespace prefix
        key = element.tag.split("}")[1] if "}" in element.tag else element.tag
        key = lower.sub("", key) or key

        # Process element as tree element if the inner XML contains non-whitespace
        # content
        if element.text and element.text.strip():
            value = element.text
        else:
            value = elem2dict(element)
        if key in result:
            if type(result[key]) is list:
                result[key].append(value)
            else:
                result[key] = [result[key], value]
        # elif key in {"no_name", "variant"}:
        #     result.update(value)
        else:
            result[key] = value

    if isinstance(result, dict):
        if "variant" in result and not isinstance(result["variant"], list):
            result = result["variant"]
        if set(result.keys()) == {"no_name"} and not isinstance(
            result["no_name"], list
        ):
            result = result["no_name"]
    # if node.tag not in {"no_name", "variant"}:
    #     result = {node.tag: result}
    return result



from __future__ import annotations

import mmap
import struct
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Dict, NewType, overload

import numpy as np
from typing_extensions import TypedDict

if TYPE_CHECKING:
    from typing import BinaryIO, Iterator, Literal, Optional, Set, Tuple, Union

    from numpy.typing import DTypeLike

# h = short              (2)
# i = int                (4)
# I = unsigned int       (4)
# Q = unsigned long long (8)
CHUNK_INFO = struct.Struct("IIQ")  # chunk_magic, shift, length
QQ = struct.Struct("QQ")
CHUNK_MAGIC = 0x0ABECEDA
CHUNK_MAP_SIGNATURE = b"ND2 CHUNK MAP SIGNATURE 0000001!"
DEFAULT_SHIFT = 4072


@contextmanager
def ensure_handle(obj: Union[str, BinaryIO]) -> Iterator[BinaryIO]:
    fh = open(obj, "rb") if isinstance(obj, (str, bytes, Path)) else obj
    try:
        yield fh
    finally:
        # close it if we were the one to open it
        if not hasattr(obj, "fileno"):
            fh.close()


FrameIndex = NewType("FrameIndex", int)
FrameOffset = NewType("FrameOffset", int)
ImageMap = Dict[FrameIndex, FrameOffset]


class FixedImageMap(TypedDict):
    bad: Set[FrameIndex]  # frames that could not be found
    fixed: Set[FrameIndex]  # frames that were bad but fixed
    # final mapping of frame number to absolute byte offset starting the chunk
    # or None, if the chunk could not be verified
    good: ImageMap


@overload
def read_chunkmap(
    file: Union[str, BinaryIO],
    *,
    validate_frames: Literal[True] = True,
    legacy: bool = False,
    search_window: int = ...,
) -> Tuple[FixedImageMap, Dict[str, int]]:
    ...


@overload
def read_chunkmap(
    file: Union[str, BinaryIO],
    *,
    validate_frames: Literal[False],
    legacy: bool = False,
    search_window: int = ...,
) -> Tuple[Dict[int, int], Dict[str, int]]:
    ...


def read_chunkmap(
    file: Union[str, BinaryIO],
    *,
    validate_frames=False,
    legacy: bool = False,
    search_window: int = 100,
):
    """Read chunkmap of nd2 `file`.

    Parameters
    ----------
    file : Union[str, BinaryIO]
        Filename or file handle to nd2 file.
    validate_frames : bool, optional
        Whether to verify (and attempt to fix) frames whose positions have been
        shifted relative to the predicted offset (i.e. in a corrupted file),
        by default False.
    legacy : bool, optional
        Treat file as legacy nd2 format, by default False
    search_window : int, optional
        When validate_frames is true, this is the search window (in KB) that will
        be used to try to find the actual chunk position. by default 100 KB

    Returns
    -------
    tuple
        (image chunk positions, metadata chunk positions).  If `validate_frames` is
        true, the image chunk dict will have three keys:
        `bad`: estimated frame positions that were invalid.
        `fixed`: estimated frame positions that were invalid, but corrected.
        `good`: estimated frame positions that were already valid.
    """
    with ensure_handle(file) as fh:
        if not legacy:
            return read_new_chunkmap(
                fh, validate_frames=validate_frames, search_window=search_window
            )
        from ._legacy import legacy_nd2_chunkmap

        d = legacy_nd2_chunkmap(fh)
        if validate_frames:
            f = {"bad": [], "fixed": [], "good": dict(enumerate(d.pop(b"LUNK")))}
            return f, d


def read_new_chunkmap(
    fh: BinaryIO, validate_frames: bool = False, search_window: int = 100
) -> Tuple[Union[ImageMap, FixedImageMap], Dict[str, int]]:
    """read the map of the chunks at the end of the file

    chunk rules:
    - each data chunk starts with
      - 4 bytes: CHUNK_MAGIC -> 0x0ABECEDA (big endian: 0xDACEBE0A)
      - 4 bytes: length of the chunk header (this section contains the chunk name...)
      - 8 bytes: length of chunk following the header, up to the next CHUNK_MAGIC
    """
    # the last 8 bytes contain the location of the beginning
    # of the chunkamp (~FILEMAP SIGNATURE NAME)
    # but we grab -40 to confirm that the CHUNK_MAP_SIGNATURE
    # string appears before the last 8 bytes.
    fh.seek(-40, 2)
    name, chunk = struct.unpack("32sQ", fh.read(40))
    assert name == CHUNK_MAP_SIGNATURE, f"Not a valid ND2 file: {fh.name}"

    # then we get all of the data in the chunkmap
    # this asserts that the chunkmap begins with CHUNK_MAGIC
    chunkmap_data = read_chunk(fh, chunk)

    # now look for each "!" in the chunkmap
    # and record the position associated with each chunkname
    pos = 0
    image_map: ImageMap = {}
    meta_map: Dict[str, int] = {}
    while True:
        # find the first "!", starting at pos, then go to next byte
        p = chunkmap_data.index(b"!", pos) + 1
        name = chunkmap_data[pos:p]  # name of the chunk
        if name == CHUNK_MAP_SIGNATURE:
            # break when we find the end
            break
        # the next 16 bytes contain...
        # (8) -> position of this key in the file  (@ the chunk magic)
        # (8) -> length of this chunk in the file (not including the chunk header)
        # Note: one still needs to go to `position` to read the CHUNK_INFO to know
        # the absolute position of the data (excluding the chunk header).  This can
        # be done using `read_chunk(..., position)``
        position, _ = QQ.unpack(chunkmap_data[p : p + 16])  # noqa
        if name[:13] == b"ImageDataSeq|":
            image_map[FrameIndex(int(name[13:-1]))] = position
        else:
            meta_map[name[:-1].decode("ascii")] = position
        pos = p + 16
    if validate_frames:
        return _validate_frames(fh, image_map, kbrange=search_window), meta_map
    image_map = {f: FrameOffset(o + 24 + DEFAULT_SHIFT) for f, o in image_map.items()}
    return image_map, meta_map


def _validate_frames(
    fh: BinaryIO, images: ImageMap, kbrange: int = 100
) -> FixedImageMap:
    """Look for invalid frames, and try to find their actual positions."""
    bad: Set[FrameIndex] = set()
    fixed: Set[FrameIndex] = set()
    good: ImageMap = {}
    _lengths = set()
    for fnum, _p in images.items():
        fh.seek(_p)
        magic, shift, length = CHUNK_INFO.unpack(fh.read(16))
        _lengths.add(length)
        if magic != CHUNK_MAGIC:  # corrupt frame
            correct_pos = _search(
                fh, b"ImageDataSeq|%a!" % fnum, images[fnum], kbrange=kbrange
            )
            if correct_pos is not None:
                fixed.add(fnum)
                good[fnum] = correct_pos + 24 + int(shift)
                images[fnum] = correct_pos
            else:
                bad.add(fnum)
        else:
            good[fnum] = FrameOffset(_p + 24 + int(shift))
    return {"bad": bad, "fixed": fixed, "good": good}


def _search(fh: BinaryIO, string: bytes, guess: int, kbrange: int = 100):
    """Search for `string`, in the `kbrange` bytes around position `guess`."""
    fh.seek(max(guess - ((1000 * kbrange) // 2), 0))
    try:
        p = fh.tell() + fh.read(1000 * kbrange).index(string) - 16
        fh.seek(p)
        if CHUNK_INFO.unpack(fh.read(CHUNK_INFO.size))[0] == CHUNK_MAGIC:
            return p
    except ValueError:
        return None


def read_chunk(handle: BinaryIO, position: int):
    handle.seek(position)
    # confirm chunk magic, seek to shift, read for length
    magic, shift, length = CHUNK_INFO.unpack(handle.read(CHUNK_INFO.size))
    assert magic == CHUNK_MAGIC, "invalid magic %x" % magic
    handle.seek(shift, 1)
    return handle.read(length)


def iter_chunks(handle) -> Iterator[Tuple[str, int, int]]:
    file_size = handle.seek(0, 2)
    handle.seek(0)
    pos = 0
    while True:
        magic, shift, length = CHUNK_INFO.unpack(handle.read(CHUNK_INFO.size))
        if magic:
            try:
                name = handle.read(shift).split(b"\x00", 1)[0].decode("utf-8")
            except UnicodeDecodeError:
                name = "?"
            yield (name, pos + +CHUNK_INFO.size + shift, length)
        pos += CHUNK_INFO.size + shift + length
        if pos >= file_size:
            break
        handle.seek(pos)


_default_chunk_start = CHUNK_MAGIC.to_bytes(4, "little")


def rescue_nd2(
    handle: Union[BinaryIO, str],
    frame_shape: Tuple[int, ...] = (),
    dtype: DTypeLike = "uint16",
    max_iters: Optional[int] = None,
    verbose=True,
    chunk_start: bytes = _default_chunk_start,
):
    """Iterator that yields all discovered frames in a file handle

    In nd2 files, each "frame" contains XY and all channel info (both true
    channels as well as RGB components).  Frames are laid out as (Y, X, C),
    and the `frame_shape` should match the expected frame size.  If
    `frame_shape` is not provided, a guess will be made about the vector shape
    of each frame, but it may be incorrect.

    Parameters
    ----------
    handle : Union[BinaryIO,str]
        Filepath string, or binary file handle (For example
        `handle = open('some.nd2', 'rb')`)
    frame_shape : Tuple[int, ...], optional
        expected shape of each frame, by default a 1 dimensional array will
        be yielded for each frame, which can be reshaped later if desired.
        NOTE: nd2 frames are generally ordered as
        (height, width, true_channels, rgbcomponents).
        So unlike numpy, which would use (channels, Y, X), you should use
        (Y, X, channels)
    dtype : np.dtype, optional
        Data type, by default np.uint16
    max_iters : Optional[int], optional
        A maximum number of frames to yield, by default will yield until the
        end of the file is reached
    verbose : bool
        whether to print info
    chunk_start

    Yields
    ------
    np.ndarray
        each discovered frame in the file

    Examples
    --------
    >>> with open('some_bad.nd2', 'rb') as fh:
    >>>     frames = rescue_nd2(fh, (512, 512, 4), 'uint16')
    >>>     ary = np.stack(frames)

    You will likely want to reshape `ary` after that.
    """
    dtype = np.dtype(dtype)
    with ensure_handle(handle) as _fh:
        mm = mmap.mmap(_fh.fileno(), 0, access=mmap.ACCESS_READ)

        offset = 0
        iters = 0
        while True:
            # search for the next part of the file starting with CHUNK_START
            offset = mm.find(chunk_start, offset)
            if offset < 0:
                if verbose:
                    print("End of file.")
                return

            # location at the end of the chunk header
            end_hdr = offset + CHUNK_INFO.size

            # find the next "!"
            # In nd2 files, each data chunk starts with the
            # string "ImageDataSeq|N" ... where N is the frame index
            next_bang = mm.find(b"!", end_hdr)
            if next_bang > 0 and (0 < next_bang - end_hdr < 128):
                # if we find the "!"... make sure we have an ImageDataSeq
                chunk_name = mm[end_hdr:next_bang]
                if chunk_name.startswith(b"ImageDataSeq|"):
                    if verbose:
                        print(f"Found image {iters} at offset {offset}")
                    # Now, read the actual data
                    _, shift, length = CHUNK_INFO.unpack(mm[offset:end_hdr])
                    # convert to numpy array and yield
                    # (can't remember why the extra 8 bytes)
                    try:
                        shape = frame_shape or ((length - 8) // dtype.itemsize,)
                        yield np.ndarray(
                            shape=shape,
                            dtype=dtype,
                            buffer=mm,
                            offset=end_hdr + shift + 8,
                        )
                    except TypeError as e:
                        # buffer is likely too small
                        if verbose:
                            print(f"Error at offset {offset}: {e}")
                    iters += 1
            elif verbose:
                print(f"Found chunk at offset {offset} with no image data")

            offset += 1
            if max_iters and iters >= max_iters:
                return



import os
import platform
from pathlib import Path

from Cython.Build import cythonize
from numpy import get_include
from setuptools import Extension, setup

SYSTEM = platform.system()
PLATFORM = platform.machine().replace("AMD64", "x86_64")
SDK = Path("src/sdk") / SYSTEM / PLATFORM
LIB = SDK / "lib"
INCLUDE = SDK / "include"
LINK = "shared" if SYSTEM == "Linux" else "static"
# set env CYTHON_TRACE=1 to enable coverage on .pyx files
CYTHON_TRACE = bool(os.getenv("CYTHON_TRACE", "0") not in ("0", "False"))

sdk = Extension(
    name="nd2._sdk.latest",
    sources=["src/nd2/_sdk/latest.pyx"],
    libraries=[f"nd2readsdk-{LINK}"],
    library_dirs=[str(LIB)],
    runtime_library_dirs=[str(LIB)] if SYSTEM == "Linux" else [],
    include_dirs=[str(INCLUDE), get_include()],
    extra_objects=[str(x) for x in LIB.glob("*") if not x.name.startswith(".")],
    define_macros=[("LX_STATIC_LINKING", None), ("CYTHON_TRACE", int(CYTHON_TRACE))],
    # extra_link_args=[
    #     "-ltiff",
    #     "-lz",
    #     "-ljpeg",
    #     "-llzma",
    #     "-ljbig",
    #     "-ltiffxx",
    #     # "-lm",
    #     # "-lstdc++fs",
    # ],
)


setup(
    use_scm_version={"write_to": "src/nd2/_version.py"},
    ext_modules=cythonize(
        [sdk],
        language_level="3",
        compiler_directives={
            "linetrace": CYTHON_TRACE,
            "c_string_type": "unicode",
            "c_string_encoding": "utf-8",
        },
    ),
)
